@@ -1,5 +1,5 @@
 #
-# Copyright © 2012 - 2020 Michal Čihař <michal@cihar.com>
+# Copyright © 2012–2022 Michal Čihař <michal@cihar.com>
 #
 # This file is part of Weblate Client <https://github.com/WeblateOrg/wlc>
 #
@@ -18,18 +18,25 @@
 #
 """Weblate API client library."""
 
+import json
+import logging
 from copy import copy
+from typing import Any, Dict, Optional, Set, Tuple
 from urllib.parse import urlencode, urlparse
 
-import dateutil.parser
+import dateutil.parser  # type: ignore
 import requests
+from requests.adapters import HTTPAdapter
+from requests.packages.urllib3.util.retry import Retry
 
-__version__ = "1.4"
+log = logging.getLogger("wlc")
+
+__version__ = "1.13"
 
 URL = "https://weblate.org/"
 DEVEL_URL = "https://github.com/WeblateOrg/wlc"
 API_URL = "http://127.0.0.1:8000/api/"
-USER_AGENT = "wlc/{0}".format(__version__)
+USER_AGENT = f"wlc/{__version__}"
 LOCALHOST_NETLOC = "127.0.0.1"
 TIMESTAMPS = {"last_change"}
 
@@ -37,37 +44,85 @@ TIMESTAMPS = {"last_change"}
 class WeblateException(Exception):
     """Generic error."""
 
+    def __init__(self, message: Optional[str] = None):
+        super().__init__(message or self.__doc__)
+
 
 class WeblateThrottlingError(WeblateException):
-    def __init__(self):
-        super().__init__("Throttling on the server")
+    """Throttling on the server."""
 
 
 class WeblatePermissionError(WeblateException):
-    def __init__(self):
-        super().__init__("You don't have permission to access this object")
+    """You don't have permission to access this object."""
 
 
 class WeblateDeniedError(WeblateException):
-    def __init__(self):
-        super().__init__("Access denied, API key is wrong or missing")
+    """Access denied, API key is wrong or missing."""
 
 
 class Weblate:
     """Weblate API wrapper object."""
 
-    def __init__(self, key="", url=API_URL, config=None):
-        """Create the object, storing key and API url."""
+    def __init__(
+        self,
+        key="",
+        url=API_URL,
+        config=None,
+        retries=0,
+        status_forcelist=None,
+        method_whitelist=None,
+        backoff_factor=0,
+        timeout=30,
+    ):
+        """Create the object, storing key, API url and requests retry args."""
+        self.session = requests.Session()
         if config is not None:
             self.url, self.key = config.get_url_key()
+            (
+                self.retries,
+                self.status_forcelist,
+                self.method_whitelist,
+                self.backoff_factor,
+                self.timeout,
+            ) = config.get_request_options()
         else:
             self.key = key
             self.url = url
+            self.retries = retries
+            self.status_forcelist = status_forcelist
+            self.timeout = timeout
+            self.method_whitelist = method_whitelist or [
+                "HEAD",
+                "GET",
+                "PUT",
+                "PATCH",
+                "DELETE",
+                "OPTIONS",
+                "TRACE",
+            ]
+            self.backoff_factor = backoff_factor
+
+        self.retries = Retry(
+            total=self.retries,
+            backoff_factor=self.backoff_factor,
+            status_forcelist=self.status_forcelist,
+            allowed_methods=self.method_whitelist,
+            raise_on_status=False,
+        )
+        self.adapter = HTTPAdapter(pool_connections=1, max_retries=retries)
+
         if not self.url.endswith("/"):
             self.url += "/"
 
     @staticmethod
-    def process_error(error):
+    def permission_error_message(error):
+        """Get detail from serialized DRF PermissionDenied exception."""
+        try:
+            return error.response.json()["detail"]
+        except (json.JSONDecodeError, KeyError):
+            return None
+
+    def process_error(self, error):
         """Raise WeblateException for known HTTP errors."""
         if isinstance(error, requests.HTTPError):
             status_code = error.response.status_code
@@ -80,64 +135,87 @@ class Weblate:
                     "(maybe operation is not supported on the server)"
                 )
             if status_code == 403:
-                raise WeblatePermissionError()
+                raise WeblatePermissionError(self.permission_error_message(error))
 
             if status_code == 401:
                 raise WeblateDeniedError()
 
             reason = error.response.reason
-            raise WeblateException("HTTP error {0}: {1}".format(status_code, reason))
+            try:
+                error_string = str(error.response.json())
+            except Exception:
+                error_string = ""
+            raise WeblateException(f"HTTP error {status_code}: {reason} {error_string}")
 
-    def raw_request(self, method, path, params=None, files=None):
+    def raw_request(self, method, path, data=None, files=None, params=None):
         """Construct request object and returns raw content."""
-        response = self.invoke_request(method, path, params, files)
+        response = self.invoke_request(method, path, data, files, params=params)
 
         return response.content
 
-    def request(self, method, path, params=None, files=None):
+    def request(self, method, path, data=None, files=None, params=None):
         """Construct request object and returns json response."""
-        response = self.invoke_request(method, path, params, files)
+        response = self.invoke_request(method, path, data, files, params=params)
 
         try:
             return response.json()
         except ValueError:
             raise WeblateException("Server returned invalid JSON")
 
-    def invoke_request(self, method, path, params=None, files=None):
+    def invoke_request(self, method, path, data=None, files=None, params=None):
         """Construct request object."""
         if not path.startswith("http"):
-            path = "{0}{1}".format(self.url, path)
+            path = f"{self.url}{path}"
         headers = {"user-agent": USER_AGENT, "Accept": "application/json"}
         if self.key:
-            headers["Authorization"] = "Token {}".format(self.key)
+            headers["Authorization"] = f"Token {self.key}"
         verify_ssl = self._should_verify_ssl(path)
+        kwargs = {
+            "headers": headers,
+            "verify": verify_ssl,
+            "files": files,
+        }
+
+        # Disable insecure warnings for localhost
+        if not verify_ssl:
+            logging.captureWarnings(True)
+        if params:
+            kwargs["params"] = params
+        if files:
+            # mulitpart/form upload
+            kwargs["data"] = data
+        else:
+            # JSON params to handle complex structures
+            kwargs["json"] = data
         try:
-            response = requests.request(
-                method,
-                path,
-                headers=headers,
-                data=params,
-                verify=verify_ssl,
-                files=files,
-            )
+            self.session.mount(f"{urlparse(path).scheme}://", self.adapter)
+            kwargs["timeout"] = self.timeout
+            log_kwargs = copy(kwargs)
+            del log_kwargs["files"]
+            log.debug(json.dumps([method, path, log_kwargs], indent=True))
+            response = self.session.request(method, path, **kwargs)
             response.raise_for_status()
         except requests.exceptions.RequestException as error:
             self.process_error(error)
             raise
         return response
 
-    def post(self, path, **kwargs):
+    def post(self, path, files=None, params=None, **kwargs):
         """Perform POST request on the API."""
-        return self.request("post", path, kwargs)
+        return self.request("post", path, data=kwargs, files=files, params=params)
 
-    def get(self, path):
+    def _post_factory(self, prefix, path, kwargs):
+        """Wrapper for posting objects."""
+        return self.post("/".join((prefix, path, "")), **kwargs)
+
+    def get(self, path, params=None):
         """Perform GET request on the API."""
-        return self.request("get", path)
+        return self.request("get", path, params=params)
 
-    def list_factory(self, path, parser):
+    def list_factory(self, path, parser, params=None):
         """Listing object wrapper."""
         while path is not None:
-            data = self.get(path)
+            data = self.get(path, params=params)
             for item in data["results"]:
                 yield parser(weblate=self, **item)
 
@@ -154,13 +232,18 @@ class Weblate:
         Operates on (project, component or translation objects.
         """
         parts = path.strip("/").split("/")
+        try:
+            int(path)
+            return self.get_unit(path)
+        except ValueError:
+            pass
         if len(parts) == 3:
             return self.get_translation(path)
         if len(parts) == 2:
             return self.get_component(path)
         if len(parts) == 1:
             return self.get_project(path)
-        raise ValueError("Not supported path: {0}".format(path))
+        raise ValueError(f"Not supported path: {path}")
 
     def get_project(self, path):
         """Return project of given path."""
@@ -174,6 +257,10 @@ class Weblate:
         """Return translation of given path."""
         return self._get_factory("translations", path, Translation)
 
+    def get_unit(self, path):
+        """Return unit of given path."""
+        return self._get_factory("units", path, Unit)
+
     def list_projects(self, path="projects/"):
         """List projects in the instance."""
         return self.list_factory(path, Project)
@@ -183,9 +270,13 @@ class Weblate:
         return self.list_factory(path, Component)
 
     def list_changes(self, path="changes/"):
-        """List components in the instance."""
+        """List changes in the instance."""
         return self.list_factory(path, Change)
 
+    def list_units(self, path, params=None):
+        """List units in the instance."""
+        return self.list_factory(path, Unit, params=params)
+
     def list_translations(self, path="translations/"):
         """List translations in the instance."""
         return self.list_factory(path, Translation)
@@ -194,6 +285,56 @@ class Weblate:
         """List languages in the instance."""
         return self.list_factory("languages/", Language)
 
+    def add_source_string(
+        self, project, component, msgid, msgstr, source_language=None
+    ):
+        """Adds a source string to a monolingual base file."""
+        if not source_language:
+            component_obj = self.get_component(f"{project}/{component}")
+            source_language = component_obj["source_language"]["code"]
+        if not isinstance(msgstr, list):
+            msgstr = [msgstr]
+        path = f"{project}/{component}/{source_language}/units"
+        payload = {"key": msgid, "value": msgstr}
+        return self._post_factory("translations", path, payload)
+
+    def create_project(
+        self, name, slug, website, source_language_name=None, source_language_code=None
+    ):
+        """Create a new project in the instance."""
+        data = {
+            "name": name,
+            "slug": slug,
+            "web": website,
+        }
+        if source_language_name and source_language_code:
+            data["source_language"] = {
+                "name": source_language_name,
+                "code": source_language_code,
+            }
+
+        return self.post("projects/", **data)
+
+    def create_component(self, project, **kwargs):
+        """Create a new component for project in the instance."""
+        files = {}
+        for fileattr in ("docfile", "zipfile"):
+            if fileattr in kwargs:
+                files[fileattr] = kwargs.pop(fileattr)
+
+        required_keys = ["name", "slug", "file_format", "filemask", "repo"]
+        for key in required_keys:
+            if key not in kwargs:
+                raise WeblateException(f"{key} is required.")
+
+        return self.post(f"projects/{project}/components/", files=files, **kwargs)
+
+    def create_language(self, code, name, direction="ltr", plural=None):
+        """Create a new language."""
+        plural = plural if plural else {"number": 2, "formula": "n != 1"}
+        data = {"code": code, "name": name, "direction": direction, "plural": plural}
+        return self.post("languages/", **data)
+
     @staticmethod
     def _should_verify_ssl(path):
         """Cheks if it should verify ssl certificates."""
@@ -205,8 +346,9 @@ class Weblate:
 class LazyObject(dict):
     """Object which supports deferred loading."""
 
-    PARAMS = ()
-    MAPPINGS = {}
+    PARAMS: Tuple[str, ...] = ()
+    OPTIONALS: Set[str] = set()
+    MAPPINGS: Dict[str, Any] = {}
     ID = "url"
 
     def __init__(self, weblate, url, **kwargs):
@@ -245,8 +387,8 @@ class LazyObject(dict):
                 else:
                     self._data[param] = value
                 del kwargs[param]
-        for key in kwargs:
-            self._attribs[key] = kwargs[key]
+        for key, value in kwargs.items():
+            self._attribs[key] = value
 
     def ensure_loaded(self, attrib):
         """Ensure attribute is loaded from remote."""
@@ -266,7 +408,10 @@ class LazyObject(dict):
             raise AttributeError(name)
         if name not in self._data:
             self.refresh()
-        return self._data[name]
+        try:
+            return self._data[name]
+        except KeyError:
+            raise AttributeError(name)
 
     def setattrvalue(self, name, value):
         if name not in self.PARAMS:
@@ -278,15 +423,20 @@ class LazyObject(dict):
         return self.__getattr__(name)
 
     def __len__(self):
-        return len(self.PARAMS)
+        return len(list(self.keys()))
 
     def keys(self):
         """Return list of attributes."""
-        return self.PARAMS
+        # There is always at least url present
+        if len(self._data) <= 1:
+            self.refresh()
+        for param in self.PARAMS:
+            if param not in self.OPTIONALS or param in self._data:
+                yield param
 
     def items(self):
         """Iterator over attributes."""
-        for key in self.PARAMS:
+        for key in self.keys():
             yield key, self.__getattr__(key)
 
     def to_value(self):
@@ -298,14 +448,14 @@ class LazyObject(dict):
 class Language(LazyObject):
     """Language object."""
 
-    PARAMS = ("url", "web_url", "code", "name", "direction")
+    PARAMS: Tuple[str, ...] = ("url", "web_url", "code", "name", "direction")
     ID = "code"
 
 
 class LanguageStats(LazyObject):
     """Language object."""
 
-    PARAMS = (
+    PARAMS: Tuple[str, ...] = (
         "total",
         "code",
         "translated_words",
@@ -349,7 +499,7 @@ class RepoMixin:
 class ProjectRepository(LazyObject, RepoMixin):
     """Repository object."""
 
-    PARAMS = ("url", "needs_commit", "needs_merge", "needs_push")
+    PARAMS: Tuple[str, ...] = ("url", "needs_commit", "needs_merge", "needs_push")
 
     def _get_repo_url(self):
         """Return repository url."""
@@ -359,7 +509,7 @@ class ProjectRepository(LazyObject, RepoMixin):
 class Repository(ProjectRepository):
     """Repository object."""
 
-    PARAMS = (
+    PARAMS: Tuple[str, ...] = (
         "url",
         "needs_commit",
         "needs_merge",
@@ -384,7 +534,15 @@ class RepoObjectMixin(RepoMixin):
 class Project(LazyObject, RepoObjectMixin):
     """Project object."""
 
-    PARAMS = ("url", "web_url", "name", "slug", "web", "source_language")
+    PARAMS: Tuple[str, ...] = (
+        "url",
+        "web_url",
+        "name",
+        "slug",
+        "web",
+        "source_language",
+    )
+    OPTIONALS = {"source_language"}
     ID = "slug"
     MAPPINGS = {"source_language": Language}
 
@@ -397,9 +555,6 @@ class Project(LazyObject, RepoObjectMixin):
         """Return statistics for translation."""
         self.ensure_loaded("statistics_url")
         data = self.weblate.get(self._attribs["statistics_url"])
-        # Changed in Weblate 4.1:
-        if "url" not in data:
-            data["url"] = self._attribs["statistics_url"]
         return Statistics(weblate=self.weblate, **data)
 
     def languages(self):
@@ -418,16 +573,20 @@ class Project(LazyObject, RepoObjectMixin):
     def delete(self):
         self.weblate.raw_request("delete", self._url)
 
+    def create_component(self, **kwargs):
+        return self.weblate.create_component(self.slug, **kwargs)
+
 
 class Component(LazyObject, RepoObjectMixin):
     """Component object."""
 
-    PARAMS = (
+    PARAMS: Tuple[str, ...] = (
         "url",
         "web_url",
         "name",
         "slug",
         "project",
+        "source_language",
         "vcs",
         "repo",
         "git_export",
@@ -438,9 +597,12 @@ class Component(LazyObject, RepoObjectMixin):
         "file_format",
         "license",
         "license_url",
+        "source_language",
+        "is_glossary",
     )
+    OPTIONALS = {"source_language", "is_glossary"}
     ID = "slug"
-    MAPPINGS = {"project": Project}
+    MAPPINGS = {"project": Project, "source_language": Language}
     REPOSITORY_CLASS = Repository
 
     def list(self):
@@ -448,6 +610,13 @@ class Component(LazyObject, RepoObjectMixin):
         self.ensure_loaded("translations_url")
         return self.weblate.list_translations(self._attribs["translations_url"])
 
+    def add_translation(self, language):
+        """Creates a new translation in the component."""
+        self.ensure_loaded("translations_url")
+        return self.weblate.post(
+            path=self._attribs["translations_url"], language_code=language
+        )
+
     def statistics(self):
         """Return statistics for component."""
         self.ensure_loaded("statistics_url")
@@ -479,11 +648,29 @@ class Component(LazyObject, RepoObjectMixin):
     def delete(self):
         self.weblate.raw_request("delete", self._url)
 
+    def add_source_string(self, msgid, msgstr):
+        """Adds a source string to a monolingual base file."""
+        return self.weblate.add_source_string(
+            project=self.project.slug,
+            component=self.slug,
+            msgid=msgid,
+            msgstr=msgstr,
+            source_language=self.source_language["code"],
+        )
+
+    def download(self, convert=None):
+        """Download translation file from server."""
+        self.ensure_loaded("repository_url")
+        url = self._get_repo_url().replace("repository", "file")
+        if convert is not None:
+            url = "{}?{}".format(url, urlencode({"format": convert}))
+        return self.weblate.raw_request("get", url)
+
 
 class Translation(LazyObject, RepoObjectMixin):
     """Translation object."""
 
-    PARAMS = (
+    PARAMS: Tuple[str, ...] = (
         "url",
         "web_url",
         "language",
@@ -523,13 +710,10 @@ class Translation(LazyObject, RepoObjectMixin):
         """Return statistics for translation."""
         self.ensure_loaded("statistics_url")
         data = self.weblate.get(self._attribs["statistics_url"])
-        # Changed in Weblate 4.1:
-        if "url" not in data:
-            data["url"] = self._attribs["statistics_url"]
         return TranslationStatistics(weblate=self.weblate, **data)
 
     def changes(self):
-        """List changes in the project."""
+        """List changes in the translation."""
         self.ensure_loaded("changes_list_url")
         return self.weblate.list_changes(self._attribs["changes_list_url"])
 
@@ -538,28 +722,35 @@ class Translation(LazyObject, RepoObjectMixin):
         self.ensure_loaded("file_url")
         url = self._attribs["file_url"]
         if convert is not None:
-            url = "{0}?{1}".format(url, urlencode({"format": convert}))
+            url = "{}?{}".format(url, urlencode({"format": convert}))
         return self.weblate.raw_request("get", url)
 
-    def upload(self, file, overwrite=None, **kwargs):
-        """Download translation file from server."""
+    def upload(self, file, overwrite=None, format=None, **kwargs):
+        """Updoad a translation file to server."""
         self.ensure_loaded("file_url")
         url = self._attribs["file_url"]
-        files = {"file": file}
-
+        if format:
+            files = {"file": (f"file.{format}", file)}
+        else:
+            files = {"file": file}
         if overwrite:
             kwargs["overwrite"] = "yes"
 
-        return self.weblate.request("post", url, files=files, params=kwargs)
+        return self.weblate.request("post", url, files=files, data=kwargs)
 
     def delete(self):
         self.weblate.raw_request("delete", self._url)
 
+    def units(self, **kwargs):
+        """List units in the translation."""
+        self.ensure_loaded("units_list_url")
+        return self.weblate.list_units(self._attribs["units_list_url"], params=kwargs)
+
 
 class Statistics(LazyObject):
     """Statistics object."""
 
-    PARAMS = (
+    PARAMS: Tuple[str, ...] = (
         "failing_percent",
         "translated_percent",
         "total_words",
@@ -572,17 +763,20 @@ class Statistics(LazyObject):
         "total",
         "last_change",
         "name",
+        "url",
     )
 
 
 class TranslationStatistics(Statistics):
-    PARAMS = Statistics.PARAMS + ("code", "last_author")
+    """Translation statistics."""
+
+    PARAMS: Tuple[str, ...] = Statistics.PARAMS + ("code", "last_author")
 
 
 class Change(LazyObject):
     """Change object."""
 
-    PARAMS = (
+    PARAMS: Tuple[str, ...] = (
         "url",
         "unit",
         "translation",
@@ -593,3 +787,52 @@ class Change(LazyObject):
     )
     ID = "id"
     MAPPINGS = {"translation": Translation, "component": Component}
+
+
+class Unit(LazyObject):
+    """Unit object."""
+
+    PARAMS: Tuple[str, ...] = (
+        "approved",
+        "content_hash",
+        "context",
+        "explanation",
+        "extra_flags",
+        "flags",
+        "fuzzy",
+        "has_comment",
+        "has_failing_check",
+        "has_suggestion",
+        "id",
+        "id_hash",
+        "location",
+        "note",
+        "num_words",
+        "position",
+        "previous_source",
+        "priority",
+        "source",
+        "source_unit",
+        "state",
+        "target",
+        "translated",
+        "translation",
+        "url",
+        "web_url",
+    )
+    ID = "id"
+    MAPPINGS = {"translation": Translation}
+
+    def list(self):
+        """API compatibility method, returns self."""
+        self.ensure_loaded("id")
+        return self
+
+    def patch(self, **kwargs):
+        return self.weblate.raw_request("patch", self._url, data=kwargs)
+
+    def put(self, **kwargs):
+        return self.weblate.raw_request("put", self._url, data=kwargs)
+
+    def delete(self):
+        return self.weblate.raw_request("delete", self._url)
