@@ -1,6 +1,6 @@
 # PYTHON_ARGCOMPLETE_OK
 #
-# Copyright © 2012 - 2020 Michal Čihař <michal@cihar.com>
+# Copyright © 2012–2022 Michal Čihař <michal@cihar.com>
 #
 # This file is part of Weblate Client <https://github.com/WeblateOrg/wlc>
 #
@@ -17,7 +17,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 #
-"""Command line interface for Weblate."""
+"""Command-line interface for Weblate."""
 import csv
 import http.client
 import json
@@ -25,18 +25,22 @@ import logging
 import sys
 from argparse import ArgumentParser
 from datetime import datetime
+from pathlib import Path
+from typing import Callable, Dict, List
 
 import argcomplete
+from requests.exceptions import RequestException
+
 import wlc
 from wlc.config import NoOptionError, WeblateConfig
 
-COMMANDS = {}
+COMMANDS: Dict[str, Callable] = {}
 
-SORT_ORDER = []
+SORT_ORDER: List[str] = []
 
 
 def register_command(command):
-    """Register command decorator in the command line interface."""
+    """Register command decorator in the command-line interface."""
     COMMANDS[command.name] = command
     return command
 
@@ -44,8 +48,8 @@ def register_command(command):
 def get_parser():
     """Create argument parser."""
     parser = ArgumentParser(
-        description="Weblate <{0}> command line utility.".format(wlc.URL),
-        epilog="This utility is developed at <{0}>.".format(wlc.DEVEL_URL),
+        description=f"Weblate <{wlc.URL}> command-line utility.",
+        epilog=f"This utility is developed at <{wlc.DEVEL_URL}>.",
     )
     parser.add_argument(
         "--format",
@@ -55,7 +59,7 @@ def get_parser():
         help="Output format to use",
     )
     parser.add_argument(
-        "--version", "-v", action="version", version="wlc {0}".format(wlc.__version__)
+        "--version", "-v", action="version", version=f"wlc {wlc.__version__}"
     )
     parser.add_argument(
         "--debug", "-D", action="store_true", help="Print verbosely http communication"
@@ -67,9 +71,12 @@ def get_parser():
     parser.add_argument("--key", "-k", help="API key")
     parser.add_argument("--url", "-u", help="API URL")
     subparser = parser.add_subparsers(
-        title="subcommands",
-        description="Subcommands specify what action to perform.",
-        dest="cmd",
+        title="Command",
+        description="""
+Specifies what action to perform.
+Invoke with --help to get more detailed help.
+    """,
+        dest="command",
     )
     subparser.required = True
 
@@ -82,7 +89,7 @@ def get_parser():
 
 
 class CommandError(Exception):
-    """Generic error from command line."""
+    """Generic error from command-line."""
 
     def __init__(self, message, detail=None):
         """Create CommandError exception."""
@@ -102,7 +109,7 @@ class DateTimeEncoder(json.JSONEncoder):
 def sort_key(value):
     """Key getter for sorting."""
     try:
-        return "{0:02d}".format(SORT_ORDER.index(value))
+        return f"{SORT_ORDER.index(value):02d}"
     except ValueError:
         return value
 
@@ -137,7 +144,7 @@ class Command:
 
     @classmethod
     def add_parser(cls, subparser):
-        """Create parser for command line."""
+        """Create parser for command-line."""
         return subparser.add_parser(cls.name, description=cls.description)
 
     def println(self, line):
@@ -152,9 +159,9 @@ class Command:
     def format_value(value):
         """Format value for rendering."""
         if isinstance(value, float):
-            return "{0:.1f}".format(value)
+            return f"{value:.1f}"
         if isinstance(value, int):
-            return "{0}".format(value)
+            return f"{value}"
         if value is None:
             return ""
         if hasattr(value, "to_value"):
@@ -180,7 +187,7 @@ class Command:
             self.println("  <thead>")
             self.println("    <tr>")
             for key in header:
-                self.println("      <th>{0}</th>".format(key))
+                self.println(f"      <th>{key}</th>")
             self.println("    </tr>")
             self.println("  </thead>")
             self.println("  <tbody>")
@@ -189,7 +196,7 @@ class Command:
                 self.println("    <tr>")
                 for key in header:
                     self.println(
-                        "      <td>{0}</td>".format(
+                        "      <td>{}</td>".format(
                             self.format_value(getattr(item, key))
                         )
                     )
@@ -200,9 +207,7 @@ class Command:
             self.println("<table>")
             for key, data in sorted_items(value):
                 self.println("  <tr>")
-                self.println(
-                    "    <th>{0}</th><td>{1}</td>".format(key, self.format_value(data))
-                )
+                self.println(f"    <th>{key}</th><td>{self.format_value(data)}</td>")
                 self.println("  </tr>")
             self.println("</table>")
 
@@ -211,13 +216,11 @@ class Command:
         if header is not None:
             for item in value:
                 for key in header:
-                    self.println(
-                        "{0}: {1}".format(key, self.format_value(getattr(item, key)))
-                    )
+                    self.println(f"{key}: {self.format_value(getattr(item, key))}")
                 self.println("")
         else:
             for key, data in sorted_items(value):
-                self.println("{0}: {1}".format(key, self.format_value(data)))
+                self.println(f"{key}: {self.format_value(data)}")
 
     def print(self, value):
         """Print value."""
@@ -246,7 +249,7 @@ class ObjectCommand(Command):
 
     @classmethod
     def add_parser(cls, subparser):
-        """Create parser for command line."""
+        """Create parser for command-line."""
         parser = super().add_parser(subparser)
         parser.add_argument(
             "object",
@@ -258,7 +261,7 @@ class ObjectCommand(Command):
         )
         return parser
 
-    def get_object(self):
+    def get_object(self, blank: bool = False):
         """Return object."""
         if self.args.object:
             path = self.args.object[0]
@@ -269,7 +272,9 @@ class ObjectCommand(Command):
                 path = None
 
         if not path:
-            raise CommandError("No object passed on command line!")
+            if blank:
+                return None
+            raise CommandError("No object passed on command-line!")
 
         return self.wlc.get_object(path)
 
@@ -287,9 +292,9 @@ class ObjectCommand(Command):
 class ProjectCommand(ObjectCommand):
     """Wrapper to allow only project objects."""
 
-    def get_object(self):
+    def get_object(self, blank: bool = False):
         """Return component object."""
-        obj = super().get_object()
+        obj = super().get_object(blank=blank)
         if not isinstance(obj, wlc.Project):
             raise CommandError("Not supported")
         return obj
@@ -302,9 +307,9 @@ class ProjectCommand(ObjectCommand):
 class ComponentCommand(ObjectCommand):
     """Wrapper to allow only component objects."""
 
-    def get_object(self):
+    def get_object(self, blank: bool = False):
         """Return component object."""
-        obj = super().get_object()
+        obj = super().get_object(blank=blank)
         if not isinstance(obj, wlc.Component):
             raise CommandError("This command is supported only at component level")
         return obj
@@ -317,9 +322,9 @@ class ComponentCommand(ObjectCommand):
 class TranslationCommand(ObjectCommand):
     """Wrapper to allow only translation objects."""
 
-    def get_object(self):
+    def get_object(self, blank: bool = False):
         """Return translation object."""
-        obj = super().get_object()
+        obj = super().get_object(blank=blank)
         if not isinstance(obj, wlc.Translation):
             raise CommandError("This command is supported only at translation level")
         return obj
@@ -338,7 +343,7 @@ class Version(Command):
 
     @classmethod
     def add_parser(cls, subparser):
-        """Create parser for command line."""
+        """Create parser for command-line."""
         parser = super().add_parser(subparser)
         parser.add_argument("--bare", action="store_true", help="Print only version")
         return parser
@@ -442,7 +447,7 @@ class Delete(ObjectCommand):
 
 
 @register_command
-class List(ObjectCommand):
+class ListObjects(ObjectCommand):
     """List object."""
 
     name = "ls"
@@ -450,10 +455,10 @@ class List(ObjectCommand):
 
     def run(self):
         """Executor."""
-        try:
-            obj = self.get_object()
+        obj = self.get_object(blank=True)
+        if obj:
             self.print(list(obj.list()))
-        except CommandError:
+        else:
             # Called without params
             lsproj = ListProjects(self.args, self.config, self.stdout)
             lsproj.run()
@@ -627,33 +632,93 @@ class Unlock(ComponentCommand):
 
 
 @register_command
-class Download(TranslationCommand):
+class Download(ObjectCommand):
     """Downloads translation file."""
 
     name = "download"
-    description = "Downloads translation file"
+    description = """
+    Downloads translation file for a project, a component or a file
+    For a project or a component we will download a zip to --output
+    """
 
     @classmethod
     def add_parser(cls, subparser):
-        """Create parser for command line."""
+        """Create parser for command-line."""
         parser = super().add_parser(subparser)
         parser.add_argument(
             "-c", "--convert", help="Convert file format on server (defaults to none)"
         )
         parser.add_argument(
-            "-o", "--output", help="File where to store output (defaults to stdout)"
+            "-o",
+            "--output",
+            help="File|Directory where to store output (defaults to stdout)",
+        )
+        parser.add_argument(
+            "-g",
+            "--no-glossary",
+            action="store_true",
+            default=False,
+            help="Disable download glossary (defaults to False)",
         )
         return parser
 
+    def download_component(self, component):
+        """Download a single component as file (if not a translation)."""
+        content = component.download(self.args.convert)
+        if self.args.output is None:
+            raise CommandError("Output is needed for download!")
+
+        directory = Path(self.args.output)
+        file_path = directory.joinpath(f"{component.project.slug}-{component.slug}.zip")
+
+        if not directory.exists():
+            directory.mkdir(exist_ok=True, parents=True)
+
+        with open(file_path, "wb") as file:
+            file.write(content)
+
+    def download_components(self, iterable):
+        for component in iterable:
+            # Ignore glossary via --no-glossary
+            if getattr(component, "is_glossary", False) and self.args.no_glossary:
+                continue
+            self.download_component(component)
+            item = f"{component.project.slug}/{component.slug}"
+            self.println(f"downloaded translations for component: {item}")
+
     def run(self):
         """Executor."""
-        obj = self.get_object()
-        content = obj.download(self.args.convert)
-        if self.args.output and self.args.output != "-":
-            with open(self.args.output, "wb") as handle:
-                handle.write(content)
-        else:
-            self.stdout.buffer.write(content)
+        obj = self.get_object(blank=True)
+
+        # Translation locale for a component
+        if isinstance(obj, wlc.Translation):
+            content = obj.download(self.args.convert)
+            if self.args.output and self.args.output != "-":
+                with open(self.args.output, "wb") as handle:
+                    handle.write(content)
+            else:
+                self.stdout.buffer.write(content)
+            return
+
+        # All translations for a component
+        if isinstance(obj, wlc.Component):
+            for component in self.wlc.list_components():
+                # Only download for the component we scoped
+                if obj.slug == component.slug:
+                    self.download_component(component)
+                    self.println(
+                        f"downloaded translations for component: {self.args.object[0]}"
+                    )
+
+            return
+
+        # All translations for a project
+        if isinstance(obj, wlc.Project):
+            self.download_components(obj.list())
+            self.println(f"downloaded translations for project: {self.args.object[0]}")
+            return
+
+        self.download_components(self.wlc.list_components())
 
 
 @register_command
@@ -665,7 +730,7 @@ class Upload(TranslationCommand):
 
     @classmethod
     def add_parser(cls, subparser):
-        """Create parser for command line."""
+        """Create parser for command-line."""
         parser = super().add_parser(subparser)
         parser.add_argument("-i", "--input", help="File to upload (defaults to stdin)")
         parser.add_argument(
@@ -679,11 +744,19 @@ class Upload(TranslationCommand):
         )
         parser.add_argument(
             "--author-email",
-            help="Author email, to override currently authenticated user",
+            help="Author e-mail, to override currently authenticated user",
         )
         parser.add_argument(
             "--method",
-            choices=("translate", "approve", "suggest", "fuzzy", "replace"),
+            choices=(
+                "translate",
+                "approve",
+                "suggest",
+                "fuzzy",
+                "replace",
+                "source",
+                "add",
+            ),
             default="translate",
         )
         parser.add_argument("--fuzzy", choices=("", "process", "approve"), default="")
@@ -720,8 +793,7 @@ class Upload(TranslationCommand):
 
 
 def parse_settings(args, settings):
-    """Read settings based on command line params."""
-
+    """Read settings based on command-line params."""
     config = WeblateConfig(args.config_section)
     if settings is None:
         config.load(args.config)
@@ -754,7 +826,7 @@ def main(settings=None, stdout=None, stdin=None, args=None):
 
     config = parse_settings(args, settings)
 
-    command = COMMANDS[args.cmd](args, config, stdout, stdin)
+    command = COMMANDS[args.command](args, config, stdout, stdin)
     try:
         command.run()
         return 0
@@ -767,12 +839,15 @@ def main(settings=None, stdout=None, stdin=None, args=None):
         else:
             print(f"Missing API key for {url}.", file=sys.stderr)
             print(
-                f"The API key can be specified by --key or in the configuration file.",
+                "The API key can be specified by --key or in the configuration file.",
                 file=sys.stderr,
             )
         return 1
+    except RequestException as error:
+        print(f"Request failed: {error}", file=sys.stderr)
+        return 10
     except (CommandError, wlc.WeblateException) as error:
-        print("Error: {0}".format(error), file=sys.stderr)
+        print(f"Error: {error}", file=sys.stderr)
         return 1
     finally:
         if args.debug:
