@@ -1,5 +1,5 @@
 #
-# Copyright © 2012 - 2020 Michal Čihař <michal@cihar.com>
+# Copyright © 2012–2022 Michal Čihař <michal@cihar.com>
 #
 # This file is part of Weblate Client <https://github.com/WeblateOrg/wlc>
 #
@@ -17,16 +17,15 @@
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 #
 """Test helpers."""
-import collections
+import cgi
 import os
-import re
+from hashlib import blake2b
+from io import BytesIO
 from unittest import TestCase
 
 import responses
 from requests.exceptions import RequestException
 
-from requests_toolbelt.multipart import decoder
-
 DATA_TEST_BASE = os.path.join(os.path.dirname(__file__), "test_data", "api")
 
 
@@ -53,55 +52,70 @@ class ResponseHandler:
         filename = self.get_filename(request)
 
         if filename is not None:
-            with open(filename, "rb") as handle:
-                return handle.read()
+            try:
+                with open(filename, "rb") as handle:
+                    return handle.read()
+            except FileNotFoundError as error:
+                error.strerror = "Failed to find response mock"
+                raise error
 
         return self.body
 
+    @staticmethod
+    def format_body(body):
+        if not body:
+            return ""
+        body = body.decode()
+        body = (
+            body.replace(": ", "=")
+            .replace("{", "")
+            .replace("}", "")
+            .replace('"', "")
+            .replace(":", "-")
+            .replace("/", "-")
+            .replace(", ", "--")
+            .replace(" ", "-")
+            .replace("[", "-")
+            .replace("]", "-")
+            .replace("*", "-")
+        )
+        return body
+
     def get_filename(self, request):
         """Return filename for given request."""
+        filename_parts = [self.filename, request.method]
         if request.method != "GET":
             content_type = request.headers.get("content-type", None)
 
             if content_type is not None and content_type.startswith(
                 "multipart/form-data"
             ):
-                return self.get_multipart_filename(content_type, request)
-            return "--".join((self.filename, request.method, request.body or ""))
+                filename_parts.append(
+                    self.format_multipart_body(request.body, content_type)
+                )
+            else:
+                filename_parts.append(self.format_body(request.body))
+            return "--".join(filename_parts)
         if "?" in request.path_url:
-            return "--".join(
-                (self.filename, request.method, request.path_url.split("?", 1)[-1])
-            )
+            filename_parts.append(request.path_url.split("?", 1)[-1])
+            return "--".join(filename_parts)
         return None
 
-    def get_multipart_filename(self, content_type, request):
-        """Return filename for given multipart request."""
-        body = request.body
-        multipart_data = decoder.MultipartDecoder(body, content_type)
-        multipart_dict = {}
-        filename_array = [self.filename, request.method]
-        for part in multipart_data.parts:
-            content_disposition = part.headers.get("Content-Disposition".encode(), None)
-
-            decoded_cd = content_disposition.decode("utf-8")
-            multipart_name = self.get_multipart_name(decoded_cd)
-
-            multipart_dict[multipart_name] = part.text.replace(" ", "-")
-
-        ordered_dict = collections.OrderedDict(sorted(multipart_dict.items()))
-
-        for key, value in ordered_dict.items():
-            filename_array.append(key + "=" + value)
-        return "--".join(filename_array)
-
-    # simple implementation instead of one based on the rfc6266 parser,
-    # as rfc6266 fails on python 3.7
     @staticmethod
-    def get_multipart_name(content_disposition):
-        """Return multipart name from content disposition."""
-        return re.search(
-            r'name\s*=\s*"(?P<name>[A-Za-z]+)"', content_disposition
-        ).group("name")
+    def format_multipart_body(body, content_type):
+        content_type, pdict = cgi.parse_header(content_type)
+        if "boundary" in pdict:
+            pdict["boundary"] = pdict["boundary"].encode()
+        fileio = BytesIO(body)
+        payload = []
+        for name, values in cgi.parse_multipart(fileio, pdict).items():
+            value = values[0]
+            if isinstance(value, bytes):
+                value = value.decode()
+            payload.append((name, value))
+        digest = blake2b(digest_size=4)
+        digest.update(repr(sorted(payload)).encode())
+        return digest.hexdigest()
 
 
 def register_uri(path, domain="http://127.0.0.1:8000/api", auth=False):
@@ -127,53 +141,73 @@ def register_uri(path, domain="http://127.0.0.1:8000/api", auth=False):
             callback=ResponseHandler(handle.read(), filename, auth),
             content_type="application/json",
         )
+        responses.add_callback(
+            responses.PATCH,
+            url,
+            callback=ResponseHandler(handle.read(), filename, auth),
+            content_type="application/json",
+        )
+        responses.add_callback(
+            responses.PUT,
+            url,
+            callback=ResponseHandler(handle.read(), filename, auth),
+            content_type="application/json",
+        )
 
 
 def raise_error(request):
     """Raise IOError."""
     if "/io" in request.path_url:
         raise RequestException("Some error")
-    raise Exception("Bug")
+    raise FileNotFoundError("Bug")
 
 
-def register_error(path, code, domain="http://127.0.0.1:8000/api", **kwargs):
+def register_error(
+    path, code, domain="http://127.0.0.1:8000/api", method=responses.GET, **kwargs
+):
     """Simplified URL error registration."""
     url = "/".join((domain, path, ""))
     if "callback" in kwargs:
-        responses.add_callback(responses.GET, url, **kwargs)
+        responses.add_callback(method, url, **kwargs)
     else:
-        responses.add(responses.GET, url, status=code, **kwargs)
+        responses.add(method, url, status=code, **kwargs)
 
 
 def register_uris():
     """Register URIs for responses."""
     paths = (
         "changes",
-        "projects",
         "components",
-        "translations",
+        "components/hello/android",
+        "components/hello/android/file",
+        "components/hello/weblate",
+        "components/hello/weblate/file",
+        "components/hello/weblate/changes",
+        "components/hello/weblate/lock",
+        "components/hello/weblate/repository",
+        "components/hello/weblate/statistics",
+        "components/hello/weblate/translations",
+        "languages",
+        "projects",
+        "projects/empty",
+        "projects/empty/components",
         "projects/hello",
         "projects/hello/changes",
         "projects/hello/components",
-        "projects/hello/statistics",
         "projects/hello/languages",
-        "projects/empty",
-        "projects/empty/components",
-        "projects/invalid",
-        "components/hello/weblate",
-        "components/hello/android",
-        "translations/hello/weblate/cs",
         "projects/hello/repository",
-        "components/hello/weblate/repository",
-        "components/hello/weblate/changes",
+        "projects/hello/statistics",
+        "projects/invalid",
+        "translations",
+        "translations/hello/weblate/cs",
+        "translations/hello/weblate/cs/changes",
         "translations/hello/weblate/cs/file",
         "translations/hello/weblate/cs/repository",
-        "translations/hello/weblate/cs/changes",
-        "components/hello/weblate/statistics",
         "translations/hello/weblate/cs/statistics",
-        "components/hello/weblate/translations",
-        "components/hello/weblate/lock",
-        "languages",
+        "translations/hello/weblate/cs/units",
+        "translations/hello/android/en/units",
+        "units",
+        "units/123",
     )
     for path in paths:
         register_uri(path)
@@ -183,6 +217,12 @@ def register_uris():
     register_uri("projects", domain="https://example.net")
     register_error("projects/nonexisting", 404)
     register_error("projects/denied", 403)
+    register_error(
+        "projects/denied_json/components",
+        403,
+        method=responses.POST,
+        json={"detail": "Can not create components"},
+    )
     register_error("projects/throttled", 429)
     register_error("projects/error", 500)
     register_error("projects/io", 500, callback=raise_error)
