@@ -1,5 +1,5 @@
 #
-# Copyright © 2012 - 2020 Michal Čihař <michal@cihar.com>
+# Copyright © 2012–2022 Michal Čihař <michal@cihar.com>
 #
 # This file is part of Weblate Client <https://github.com/WeblateOrg/wlc>
 #
@@ -16,15 +16,13 @@
 # You should have received a copy of the GNU General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
 #
-"""Test command line interface."""
+"""Test command-line interface."""
 
 import json
 import os
 import sys
 from io import BytesIO, StringIO, TextIOWrapper
-from tempfile import NamedTemporaryFile
-
-from requests.exceptions import RequestException
+from tempfile import NamedTemporaryFile, TemporaryDirectory
 
 import wlc
 from wlc.config import WeblateConfig
@@ -32,8 +30,9 @@ from wlc.main import Version, main
 
 from .test_base import APITest
 
-TEST_CONFIG = os.path.join(os.path.dirname(__file__), "test_data", "wlc")
-TEST_SECTION = os.path.join(os.path.dirname(__file__), "test_data", "section")
+TEST_DATA = os.path.join(os.path.dirname(__file__), "test_data")
+TEST_CONFIG = os.path.join(TEST_DATA, "wlc")
+TEST_SECTION = os.path.join(TEST_DATA, "section")
 
 
 def execute(args, settings=None, stdout=None, stdin=None, expected=0):
@@ -66,7 +65,7 @@ class TestSettings(APITest):
     """Test settings handling."""
 
     def test_commandline(self):
-        """Configuration using commandline."""
+        """Configuration using command-line."""
         output = execute(["--url", "https://example.net/", "list-projects"])
         self.assertIn("Hello", output)
 
@@ -108,6 +107,16 @@ class TestSettings(APITest):
         )
         self.assertIn("ACL", output)
 
+    def test_config_appdata(self):
+        """Configuration using custom config file section and key set."""
+        output = execute(["show", "acl"], settings=False, expected=1)
+        try:
+            os.environ["APPDATA"] = TEST_DATA
+            output = execute(["show", "acl"], settings=False)
+            self.assertIn("ACL", output)
+        finally:
+            del os.environ["APPDATA"]
+
     def test_config_cwd(self):
         """Test loading settings from current dir."""
         current = os.path.abspath(".")
@@ -118,6 +127,19 @@ class TestSettings(APITest):
         finally:
             os.chdir(current)
 
+    def test_default_config_values(self):
+        """Test default parser values."""
+        config = WeblateConfig()
+        self.assertEqual(config.get("weblate", "key"), "")
+        self.assertEqual(config.get("weblate", "retries"), 0)
+        self.assertEqual(config.get("weblate", "timeout"), 30)
+        self.assertEqual(
+            config.get("weblate", "method_whitelist"),
+            "HEAD\nTRACE\nDELETE\nOPTIONS\nPUT\nGET",
+        )
+        self.assertEqual(config.get("weblate", "backoff_factor"), 0)
+        self.assertEqual(config.get("weblate", "status_forcelist"), None)
+
     def test_parsing(self):
         """Test config file parsing."""
         config = WeblateConfig()
@@ -125,6 +147,29 @@ class TestSettings(APITest):
         config.load()
         config.load(TEST_CONFIG)
         self.assertEqual(config.get("weblate", "url"), "https://example.net/")
+        self.assertEqual(config.get("weblate", "retries"), "999")
+        self.assertEqual(config.get("weblate", "method_whitelist"), "PUT,POST")
+        self.assertEqual(config.get("weblate", "backoff_factor"), "0.2")
+        self.assertEqual(
+            config.get("weblate", "status_forcelist"), "429,500,502,503,504"
+        )
+
+    def test_get_request_options(self):
+        """Test the get_request_options method when all options are in config."""
+        config = WeblateConfig()
+        config.load()
+        config.load(TEST_CONFIG)
+        (
+            retries,
+            status_forcelist,
+            method_whitelist,
+            backoff_factor,
+            _timeout,
+        ) = config.get_request_options()
+        self.assertEqual(retries, 999)
+        self.assertEqual(status_forcelist, [429, 500, 502, 503, 504])
+        self.assertEqual(method_whitelist, ["PUT", "POST"])
+        self.assertEqual(backoff_factor, 0.2)
 
     def test_argv(self):
         """Test sys.argv processing."""
@@ -132,7 +177,7 @@ class TestSettings(APITest):
         try:
             sys.argv = ["wlc", "version"]
             output = execute(None)
-            self.assertIn("version: {0}".format(wlc.__version__), output)
+            self.assertIn(f"version: {wlc.__version__}", output)
         finally:
             sys.argv = backup
 
@@ -143,7 +188,7 @@ class TestOutput(APITest):
     def test_version_text(self):
         """Test version printing."""
         output = execute(["--format", "text", "version"])
-        self.assertIn("version: {0}".format(wlc.__version__), output)
+        self.assertIn(f"version: {wlc.__version__}", output)
 
     def test_version_json(self):
         """Test version printing."""
@@ -154,7 +199,7 @@ class TestOutput(APITest):
     def test_version_csv(self):
         """Test version printing."""
         output = execute(["--format", "csv", "version"])
-        self.assertIn("version,{0}".format(wlc.__version__), output)
+        self.assertIn(f"version,{wlc.__version__}", output)
 
     def test_version_html(self):
         """Test version printing."""
@@ -164,7 +209,7 @@ class TestOutput(APITest):
     def test_projects_text(self):
         """Test projects printing."""
         output = execute(["--format", "text", "list-projects"])
-        self.assertIn("name: {0}".format("Hello"), output)
+        self.assertIn("name: {}".format("Hello"), output)
 
     def test_projects_json(self):
         """Test projects printing."""
@@ -196,7 +241,7 @@ class TestCommands(APITest):
     def test_version_bare(self):
         """Test version printing."""
         output = execute(["version", "--bare"])
-        self.assertEqual("{0}\n".format(wlc.__version__), output)
+        self.assertEqual(f"{wlc.__version__}\n", output)
 
     def test_ls(self):
         """Project listing."""
@@ -254,9 +299,8 @@ class TestCommands(APITest):
         self.assertIn("/hello/weblate/cs/", output)
 
     def test_show_error(self):
-        with self.assertRaises(RequestException):
-            execute(["show", "io"])
-        with self.assertRaises(Exception):
+        execute(["show", "io"], expected=10)
+        with self.assertRaises(FileNotFoundError):
             execute(["show", "bug"])
 
     def test_delete(self):
@@ -376,6 +420,12 @@ class TestCommands(APITest):
 
     def test_download(self):
         """Translation file downloads."""
+        output = execute(["download"], expected=1)
+        self.assertIn("Output is needed", output)
+
+        with TemporaryDirectory() as tmpdirname:
+            execute(["download", "--output", tmpdirname])
+
         output = execute(["download", "hello/weblate/cs"])
         self.assertIn(b"Plural-Forms:", output)
 
@@ -390,12 +440,41 @@ class TestCommands(APITest):
             self.assertIn(b"Plural-Forms:", output)
 
         output = execute(["download", "hello/weblate"], expected=1)
-        self.assertIn("This command is supported only at translation level", output)
+        self.assertIn("Output is needed", output)
+
+        with TemporaryDirectory() as tmpdirname:
+            execute(["download", "hello/weblate", "--output", tmpdirname])
+
+        output = execute(["download", "hello"], expected=1)
+        self.assertIn("Output is needed", output)
+
+        with TemporaryDirectory() as tmpdirname:
+            execute(["download", "hello", "--no-glossary", "--output", tmpdirname])
+            # The hello-android should not be present as it is flagged as a glossary
+            self.assertEqual(os.listdir(tmpdirname), ["hello-weblate.zip"])
+
+        with TemporaryDirectory() as tmpdirname:
+            execute(
+                [
+                    "download",
+                    "hello",
+                    "--convert",
+                    "zip",
+                    "--output",
+                    os.path.join(tmpdirname, "output"),
+                ]
+            )
+            self.assertEqual(os.listdir(tmpdirname), ["output"])
 
     def test_upload(self):
         """Translation file uploads."""
         msg = "Error: Failed to upload translations!\nNot found.\n"
 
+        output = execute(["upload", "hello/weblate"], expected=1)
+        self.assertEqual(
+            "Error: This command is supported only at translation level\n", output
+        )
+
         with self.get_text_io_wrapper("test upload data") as stdin:
             output = execute(["upload", "hello/weblate/cs"], stdin=stdin)
             self.assertEqual("", output)
@@ -405,7 +484,7 @@ class TestCommands(APITest):
             self.assertEqual(msg, output)
 
         with NamedTemporaryFile(delete=False) as handle:
-            handle.write("test upload overwrite".encode())
+            handle.write(b"test upload overwrite")
             handle.close()
             output = execute(
                 ["upload", "hello/weblate/cs", "-i", handle.name, "--overwrite"]
@@ -419,15 +498,17 @@ class TestCommands(APITest):
 
 
 class TestErrors(APITest):
+    """Error handling tests."""
+
     def test_commandline_missing_key(self):
-        """Configuration using commandline."""
+        """Configuration using command-line."""
         output = execute(
             ["--url", "http://denied.example.com", "list-projects"], expected=1
         )
         self.assertIn("Missing API key", output)
 
     def test_commandline_wrong_key(self):
-        """Configuration using commandline."""
+        """Configuration using command-line."""
         output = execute(
             ["--key", "x", "--url", "http://denied.example.com", "list-projects"],
             expected=1,
